#![allow(dead_code)]

use glam::Vec2;
use serde::{Deserialize, Serialize};

#[repr(C)]
#[derive(Default, Clone, Copy, Debug, PartialEq, Serialize, Deserialize)]
pub struct Rect {
	pub min: Vec2,
	pub max: Vec2,
}

impl Rect {
	pub const EMPTY: Self = Self {
		max: Vec2::NEG_INFINITY,
		min: Vec2::INFINITY,
	};
	#[inline]
	pub fn new(x0: f32, y0: f32, x1: f32, y1: f32) -> Self {
		Self::from_corners(Vec2::new(x0, y0), Vec2::new(x1, y1))
	}

	#[inline]
	pub fn from_corners(p0: Vec2, p1: Vec2) -> Self {
		Self {
			min: p0.min(p1),
			max: p0.max(p1),
		}
	}

	#[inline]
	pub fn from_center_size(origin: Vec2, size: Vec2) -> Self {
		assert!(size.cmpge(Vec2::ZERO).all(), "Rect size must be positive");
		let half_size = size / 2.;
		Self::from_center_half_size(origin, half_size)
	}

	#[inline]
	pub fn from_center_half_size(origin: Vec2, half_size: Vec2) -> Self {
		assert!(
			half_size.cmpge(Vec2::ZERO).all(),
			"Rect half_size must be positive"
		);
		Self {
			min: origin - half_size,
			max: origin + half_size,
		}
	}

	#[inline]
	pub fn is_empty(&self) -> bool {
		self.min.cmpge(self.max).any()
	}

	#[inline]
	pub fn width(&self) -> f32 {
		self.max.x - self.min.x
	}

	#[inline]
	pub fn height(&self) -> f32 {
		self.max.y - self.min.y
	}

	#[inline]
	pub fn size(&self) -> Vec2 {
		self.max - self.min
	}

	#[inline]
	pub fn half_size(&self) -> Vec2 {
		self.size() * 0.5
	}

	#[inline]
	pub fn center(&self) -> Vec2 {
		(self.min + self.max) * 0.5
	}

	#[inline]
	pub fn contains(&self, point: Vec2) -> bool {
		(point.cmpge(self.min) & point.cmple(self.max)).all()
	}

	#[inline]
	pub fn union(&self, other: Self) -> Self {
		Self {
			min: self.min.min(other.min),
			max: self.max.max(other.max),
		}
	}

	#[inline]
	pub fn union_point(&self, other: Vec2) -> Self {
		Self {
			min: self.min.min(other),
			max: self.max.max(other),
		}
	}

	#[inline]
	pub fn intersect(&self, other: Self) -> Self {
		let mut r = Self {
			min: self.min.max(other.min),
			max: self.max.min(other.max),
		};
		// Collapse min over max to enforce invariants and ensure e.g. width() or
		// height() never return a negative value.
		r.min = r.min.min(r.max);
		r
	}

	#[inline]
	pub fn inflate(&self, expansion: f32) -> Self {
		let mut r = Self {
			min: self.min - expansion,
			max: self.max + expansion,
		};
		// Collapse min over max to enforce invariants and ensure e.g. width() or
		// height() never return a negative value.
		r.min = r.min.min(r.max);
		r
	}

	pub fn normalize(&self, other: Self) -> Self {
		let outer_size = other.size();
		Self {
			min: (self.min - other.min) / outer_size,
			max: (self.max - other.min) / outer_size,
		}
	}
}
